
#define SMBUS_BASE 0x5000
#define TVX_VT1621_PORT 0x40

#include "modedb.h"
#include "tvout.h"
#include "tvout_io.h"

int cyberblade_crtc=0x3d0;
#define CRINB(reg) (IO_OUTB((reg),(cyberblade_crtc+4)),IO_INB((cyberblade_crtc+5)))
#define CROUTB(reg,val) (IO_OUTB((reg),(cyberblade_crtc+4)),IO_OUTB((val),(cyberblade_crtc+5)))
#define SRINB(reg) (IO_OUTB((reg),0x3c4),IO_INB(0x3c5))
#define SROUTB(reg,val) (IO_OUTB((reg),0x3c4),IO_OUTB((val),0x3c5))

/* SMBus routines borrowed from VIA's Trident Driver */
static unsigned char smbus_read(unsigned char bIndex, unsigned char devAdr)
{
    unsigned short i;
    unsigned char  bData;

    /* clear host status */
    IO_OUTB(0xff, SMBUS_BASE);

    /* check SMBUS ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (IO_INB(SMBUS_BASE) & 0x01) == 0 )
            break;

    /* set host command */
    IO_OUTB(bIndex, SMBUS_BASE+3);

    /* set slave address */
    IO_OUTB(devAdr | 0x01, SMBUS_BASE+4);

    /* start */
    IO_OUTB(0x48, SMBUS_BASE+2);

    /* SMBUS Wait Ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (IO_INB(SMBUS_BASE) & 0x01) == 0 )
            break;
    bData=IO_INB(SMBUS_BASE+5);

    return bData;

}


static void smbus_write(unsigned char bData, unsigned char bIndex, unsigned char devAdr)
{
    unsigned short i;

    /* clear host status */
    IO_OUTB(0xff, SMBUS_BASE);

    /* check SMBUS ready */
    for ( i = 0; i < 0xFFFF; i++ )
        if ( (IO_INB(SMBUS_BASE) & 0x01) == 0 )
           break;

    IO_OUTB(0x08, SMBUS_BASE+2);

    /* set host command */
    IO_OUTB(bIndex, SMBUS_BASE+3);

    /* set slave address */
    IO_OUTB(devAdr & 0xFE, SMBUS_BASE+4);

    IO_OUTB(bData, SMBUS_BASE+5);

    /* start */
    IO_OUTB(0x48, SMBUS_BASE+2);

    /* SMBUS Wait Ready */
    for ( i = 0; i < 0xFFFF; i++ )
    if ( (IO_INB(SMBUS_BASE) & 0x01) == 0 )
       break;
}

/* Set/Clear register protection in SR11 */

static void SetProtect(int val)
{
	static int refcount=0,oldval=0;
	if(val)
	{
		if(refcount)
			--refcount;
		else
			SROUTB(0x11,oldval);
	}
	else
	{
		if(refcount)
			++refcount;
		else
			oldval=SRINB(0x11);
		SROUTB(0x11,0x92);
	}
}

/* --- */

int TV_InitHardware(void)
{
	IO_INIT
	if(IO_INB(0x3cc)&1)
		cyberblade_crtc=0x3d0;
	else
		cyberblade_crtc=0x3b0;
	return 0;
}


static void NearestClock(double target,int *dest_n,int *dest_m,int *dest_k)
{
	int n,m,k,bn=0,bm=0,bk=0;
	double ref=REFERENCE_CLOCK,t,bt=0,d,bd=1000.0;
	for(k=1;k<4;++k)
	{
		for(m=1;m<64;++m)
		{
			double pm=m+2.0;
			int fn=(pm*(1<<k)*target)/ref - 8.5;
			for(n=fn;n<256;++n)
			{
				t=(ref*(n+8.0)) / (pm * (1<<k));
				d=target-t; if(d<0) d=-d;
				if(d<bd)
				{
					bd=d; bt=t;
					bn=n; bm=m; bk=k;
				}
				else
				{
					if (t>target)
						n=256;
				}
			}
		}
	}
	*dest_n=bn;
	*dest_m=bm;
	*dest_k=bk;
}


static int IsTrueColour()
{
	int pixfmt=CRINB(0x38);
	if(pixfmt&0x28)
		return(1);
	else
		return(0);
}


static void SetTVCRTC(struct TVOutMode *mode)
{
	int n,m,k,vrtstart;
	int overflow;
	NearestClock(mode->Clock,&n,&m,&k);
	if(IsTrueColour())
		k-=1; /* 32 bit modes need twice the clock frequency... */

	if(mode->TVStd==TVSTD_NTSC)
		CROUTB(0xc0,0x02);
	else
		CROUTB(0xc0,0x82);

	CROUTB(0xd0,mode->HData);
	CROUTB(0xd1,0xe0); /* VGA Slave, H/V Data Load */
	CROUTB(0xd2,mode->VData&0xff);
	CROUTB(0xd3,(mode->VData>>8)&0xff);

	CROUTB(0xe0,((mode->HTotal-40)>>3));
	CROUTB(0xe3,0x00); /* HBEnd */

	CROUTB(0xe4,(mode->HDisp+48)>>3);
	CROUTB(0xe5,(mode->HDisp+56)>>3);
	CROUTB(0xe6,(mode->VTotal-2));
	vrtstart=(mode->VTotal+mode->VDisp)/2;
	overflow=((mode->VTotal-2)>>8)&0x01;
	overflow|=((mode->VTotal-2)>>4)&0x20;
	overflow|=(vrtstart>>6)&0x04;
	overflow|=(vrtstart>>2)&0x80;
	CROUTB(0xe7,overflow);

	CROUTB(0xf0,vrtstart);
	CROUTB(0xf1,4);
	CROUTB(0xf6,mode->VTotal-3); /* VBEnd */

	CROUTB(0xfe,n);
	CROUTB(0xff,k<<6|m);

}


static void SetTVX(struct TVOutMode *mode)
{
	unsigned long fsci=CalcFSCI(mode->TVStd);
	int tmp;
	int std=mode->TVStd;
	if(std>0) std+=1;

	smbus_write((mode->FrameRes<<5)|0x0b,0x00,TVX_VT1621_PORT);
	smbus_write(0x0b,0x01,TVX_VT1621_PORT); /* VSync output, H/VSync Active high */
	smbus_write((mode->Scale<<3)|0x03,0x02,TVX_VT1621_PORT);
	smbus_write(0x02,0x03,TVX_VT1621_PORT); /* Luma DeFlicker 1:2:1 */
	smbus_write((std)|0x40,0x04,TVX_VT1621_PORT); /* TV Standard + YCC to YUV conversion */

	smbus_write(0x50,0x05,TVX_VT1621_PORT); /* Slave Clock, Auto FSCI Adjust */
	smbus_write(0x12,0x06,TVX_VT1621_PORT); /* 2x Input Clock, Clock Adjust 2 */

	tmp=mode->StartAV;
	if(!(IsTrueColour()))
		tmp+=1;
	smbus_write(tmp,0x7,TVX_VT1621_PORT);
	smbus_write(mode->StartH,0x8,TVX_VT1621_PORT);
	smbus_write(mode->StartV,0x9,TVX_VT1621_PORT);

	tmp=(mode->StartAV>>5)&0x8;
	tmp|=(mode->StartH>>6)&0x4;
	tmp|=(mode->StartV>>7)&0x2;
	smbus_write(tmp,0x1c,TVX_VT1621_PORT);

	smbus_write(0xe0,0x0a,TVX_VT1621_PORT); /* Cr Amplitude */
	smbus_write(0x00,0x0b,TVX_VT1621_PORT); /* Black level */
	smbus_write(0x80,0x0c,TVX_VT1621_PORT); /* Luma Amplitude */
	smbus_write(0xe0,0x0d,TVX_VT1621_PORT); /* Cb Amplitude */
	smbus_write(0x00,0x0e,TVX_VT1621_PORT); /* Power management, all off */
	smbus_write(0x81,0x0f,TVX_VT1621_PORT); /* Macrovision, CVBST on */
	smbus_write(0x00,0x10,TVX_VT1621_PORT); /* Hue adjust */
	smbus_write(0x00,0x11,TVX_VT1621_PORT); /* Hue adjust (top three), bit 3 dot crawl */

	smbus_write(mode->PLL_P2&0xf,0x12,TVX_VT1621_PORT);
	smbus_write(mode->PLL_D,0x13,TVX_VT1621_PORT);
	smbus_write(mode->PLL_N&0xff,0x14,TVX_VT1621_PORT);
	smbus_write(((mode->PLL_P1&0xf)<<2) | ((mode->PLL_N>>7)&0x02),0x15,TVX_VT1621_PORT);

	smbus_write(fsci&0xff,0x16,TVX_VT1621_PORT); fsci>>=8;
	smbus_write(fsci&0xff,0x17,TVX_VT1621_PORT); fsci>>=8;
	smbus_write(fsci&0xff,0x18,TVX_VT1621_PORT); fsci>>=8;
	smbus_write(fsci&0xff,0x19,TVX_VT1621_PORT);

        smbus_write(0x0c,0x1a,TVX_VT1621_PORT); 

	smbus_write(0x80,0x1d,TVX_VT1621_PORT); /* Tests off */
	smbus_write(0x00,0x1e,TVX_VT1621_PORT); /* Tests off */
	smbus_write(0x00,0x1f,TVX_VT1621_PORT); /* Tests off */

        smbus_write(0x0a,0x20,TVX_VT1621_PORT); /* TV Sync Step */
        smbus_write(0x05,0x21,TVX_VT1621_PORT); /* TV Burst Envelope Step */  
        smbus_write(0x5e,0x22,TVX_VT1621_PORT); /* TV Sub-carrier Phase Adjustment */  
        smbus_write(0x83,0x23,TVX_VT1621_PORT); /* TV Blank level */  
        smbus_write(0x08,0x24,TVX_VT1621_PORT); /* TV Signal overflow */
  
        smbus_write(0xa0,0x4a,TVX_VT1621_PORT); /* Input data threshold */  
        smbus_write(0x22,0x4b,TVX_VT1621_PORT); /* Input aperture delta */  
        smbus_write(0x00,0x4c,TVX_VT1621_PORT); /* Text enhancement upper threshold */  
        smbus_write(0x00,0x4d,TVX_VT1621_PORT); /* Text enhancement lower threshold */  
        smbus_write(0x00,0x4e,TVX_VT1621_PORT); /* Aperture mode */  
        smbus_write(0x00,0x4f,TVX_VT1621_PORT); /* Coring function */  
        smbus_write(0x00,0x50,TVX_VT1621_PORT); /* Y Delay control */  
        smbus_write(0x0c,0x51,TVX_VT1621_PORT); /* YCC Underflow check, V Delay depth 4 */  
        smbus_write(0x4d,0x52,TVX_VT1621_PORT); /* Burst maximum amplitude */  

        smbus_write(0x0a,0x5a,TVX_VT1621_PORT); /* Undocumented */  
        smbus_write(0x74,0x5b,TVX_VT1621_PORT); /* Undocumented */  
        smbus_write(0x10,0x60,TVX_VT1621_PORT); /* Undocumented */  
        smbus_write(0x6f,0x61,TVX_VT1621_PORT); /* Undocumented */  
}


void SetMisc(struct TVOutMode *mode)
{
	SROUTB(0x24,0x4f);
	CROUTB(0xc1,0x41);
	IO_OUTB(0x23,0x3ce);
	IO_OUTB(0x88,0x3cf);

	/* set CRT + TV */
	IO_OUTB(0x33,0x3ce);
	IO_OUTB(0x20,0x3cf);

	SROUTB(0xD8,0x60);
	SROUTB(0xD9,0x38);
}


struct TVOutMode *GuessTVMode()
{
	if(TVOutActive())
	{
		int htotal,vtotal,frameres,tvstd;
		htotal=CRINB(0xe0)<<3;
		htotal+=40;
		vtotal=CRINB(0xe6);
		vtotal|=(CRINB(0xe7)&1)<<8;
		vtotal|=(CRINB(0xe7)&0x20)<<4;
		vtotal+=2;

		frameres=smbus_read(0x00,TVX_VT1621_PORT)>>5;
		tvstd=smbus_read(0x04,TVX_VT1621_PORT)&0x3;
		if(tvstd>0)
			tvstd-=1;

		return(MatchTVMode(tvstd,frameres,OS_NULL,SCALE_NULL,-1,-1,htotal,vtotal));
	}
	return(0);
}


void GetCurrentResolution(int *hdisp,int *vdisp)
{
	int h,v;
	h=(CRINB(0x01)+1)*8; /* H Display Enable End */
	v=CRINB(0x12); /* V Display Enable End */
	if(CRINB(0x07)&0x02) v|=256;
	if(CRINB(0x07)&0x40) v|=512;
	v+=1;
	*hdisp=h;
	*vdisp=v;
}

void SetTVMode(struct TVOutMode *mode)
{
	SetProtect(0);
	SetMisc(mode);
	SetTVCRTC(mode);
	SetTVX(mode);
	SetProtect(1);
}


int TVOutActive()
{
	int val=CRINB(0xd1);
	return((val&0x80)!=0);
}

void TVOutOff()
{
	SetProtect(0);

	CROUTB(0xc1,0x00);
	CROUTB(0xd1,0x00);
        SROUTB(0x24,0x4f); 
        SROUTB(0xD9,0x18);  

	IO_OUTB(0x30,0x3ce);
	IO_OUTB(0x00,0x3cf);

        IO_OUTB(0x33,0x3ce);
        IO_OUTB(0x20,0x3cf);

	IO_OUTB(0x23,0x3ce);
	IO_OUTB(0x88,0x3cf);
	
	SetProtect(1);
}
