/*
   TVTool 0.0.2 by Alastair M. Robinson
   blackfive@fakenhamweb.co.uk
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "tvout.h"
#include "modedb.h"

struct tvtool_options
{
	int TVStandard;
	int Overscan;
	int FrameRes;
	int TVActive;
};

struct tvtool_options options=
{
	-1,
	-1,
	-1,
	1,
};

enum option_result {OPT_FAIL,OPT_PASSIVE,OPT_ACTIVE};

int process_options(int argc,char *argv[])
{
	int result=OPT_PASSIVE;
	while(1)
	{
		int c;
		static struct option long_options[] =
		{
			{"help",no_argument,NULL,'h'},
			{"version",no_argument,NULL,'v'},
			{"overscan",required_argument,NULL,'o'},
			{"tvstd",required_argument,NULL,'s'},
			{"tvactive",required_argument,NULL,'t'},
			{0, 0, 0, 0}
		};
		c = getopt_long(argc,argv,"hvo:s:t:",long_options,NULL);
		if(c==-1)
			break;
		switch (c)
		{
			case 'h':
				printf("Usage: %s [options]\n",argv[0]);
				printf("\t -h --help\t\tdisplay this message\n");
				printf("\t -v --version\t\tdisplay version\n");
				printf("\t -o --overscan\t\tsets the overscan type, if possible.\n");
				printf("\t\t\t\tPossible values are:\n");
				printf("\t\t\t\t\t0: Normal (default)\n");
				printf("\t\t\t\t\t1: Overscan\n");
				printf("\t\t\t\t\t2: Underscan\n");
				printf("\t\t\t\tNot all settings are available in all resolutions.\n");
				printf("\t -s --tvstd\t\tsets the TV Standard.\n");
				printf("\t\t\t\tPossible values are:\n");
				printf("\t\t\t\t\t0: PAL (default)\n");
				printf("\t\t\t\t\t1: PAL-M\n");
				printf("\t\t\t\t\t2: NTSC\n");
				printf("\t\t\t\tWARNING: PAL-M mode is not yet fully tested\n");
				printf("\t -t --tvactive\t\t[Experimental] turns tv-out on (1) or off (0)\n");
				return(OPT_FAIL);
			case 'v':
				printf("tvtool-0.1 for VIA EPIA Motherboard\n");
				return(OPT_FAIL);
			case 'o':
				if(optarg)
				options.Overscan=atoi(optarg);
				if(options.Overscan>=OVERSCAN_COUNT)
				{
				  printf("Error: Overscan must be 0, 1 or 2\n");
					return(OPT_FAIL);
				}
				result=OPT_ACTIVE;
				break;
			case 's':
				if(optarg)
				options.TVStandard=atoi(optarg);
				if(options.TVStandard>=TVSTANDARD_COUNT)
				{
					printf("Error: TVStandard must be 0,1 or 2\n");
					return(OPT_FAIL);
				}
				result=OPT_ACTIVE;
				break;
			case 't':
				if(optarg)
				options.TVActive=atoi(optarg);
				if(options.TVActive>1)
				{
					printf("Error: TVActive must be 0 or 1\n");
					return(OPT_FAIL);
				}
				result=OPT_ACTIVE;
				break;
		}
	}
	if(optind < argc-1)
	{
		fprintf(stderr,"%s: too many arguments\n",argv[0]);
		return(OPT_FAIL);
	}
	return(result);
}


int main(int argc, char *argv[])
{
	int action;
	if((action=process_options(argc,argv)))
	{
		struct TVOutMode *mode=NULL;
		int hr=-1,vr=-1;

		TV_InitHardware();

		if(options.TVActive==0)
		{
			TVOutOff();
			return(0);
		}
		
		GetCurrentResolution(&hr,&vr);

		if(TVOutActive())
		{
			printf("TV-Out active\n");
			if((mode=GuessTVMode()))
			{
				printf("Current TV mode:\n");
				printf("  TVStandard: %s\n",TVStdText(mode->TVStd));
				printf("  TVX Resolution: %s\n",FrameResText(mode->FrameRes));
				printf("  Actual Resolution: %dx%d\n",hr,vr);
				printf("  Size: %s (scale %s)\n",OverscanText(mode->Overscan),ScaleText(mode->Scale));

			}
			else
				printf("Current TV mode unknown\n");
		}
		else
		{
			printf("TVOut currently inactive\n");
		}

		if(action==OPT_ACTIVE)
		{
			if(mode)
			{
				if(options.TVStandard==-1) options.TVStandard=mode->TVStd;
				if(options.Overscan==-1) options.Overscan=mode->Overscan;
				if(options.FrameRes==-1) options.FrameRes=mode->FrameRes;
			}
			else
			{
				printf("Current resolution: %dx%d\n",hr,vr);
			}

			if((mode=MatchTVMode(options.TVStandard,-1,options.Overscan,-1,hr,vr,-1,-1)))
			{
				printf("Changing mode to:\n");
				printf("  TVStandard: %s\n",TVStdText(mode->TVStd));
				printf("  Resolution: %s\n",FrameResText(mode->FrameRes));
				printf("  Size: %s (scale %s)\n",OverscanText(mode->Overscan),ScaleText(mode->Scale));
				SetTVMode(mode);
			}
			else if(options.Overscan>-1)
			{
				printf("%s mode not available - matching without Overscan...\n",OverscanText(options.Overscan));
				if((mode=MatchTVMode(options.TVStandard,-1,-1,hr,vr,-1,-1,-1)))
				{
					printf("Changing mode to:\n");
					printf("  TVStandard: %s\n",TVStdText(mode->TVStd));
					printf("  Resolution: %s\n",FrameResText(mode->FrameRes));
					printf("  Size: %s (scale %s)\n",OverscanText(mode->Overscan),ScaleText(mode->Scale));
					SetTVMode(mode);
				}
				else
					printf("Error: No suitable mode available\n");
			}
			else
				printf("Error: No suitable mode available\n");
		}
	}
	return(0);
}
